@@ -21,7 +21,6 @@ struct thread_buf {
 	char buf[THREAD_BUF_SIZE];
 };
 
-static int version;
 struct rtpengine_common_config *rtpe_common_config_ptr;
 __thread struct timeval rtpe_now;
 volatile bool rtpe_shutdown;
@@ -48,7 +47,7 @@ void daemonize(void) {
 	setpgrp();
 }
 
-void wpidfile() {
+void wpidfile(void) {
 	FILE *fp;
 
 	if (!rtpe_common_config_ptr->pidfile)
@@ -153,34 +152,20 @@ void config_load_free(struct rtpengine_common_config *cconfig) {
 	g_free(cconfig->pidfile);
 }
 
-static void free_gkeyfile(GKeyFile **k) {
-	if (k && *k)
-		g_key_file_free(*k);
-}
-static void free_gopte(GOptionEntry **k) {
-	if (k && *k)
-		free(*k);
-}
-static void free_goptc(GOptionContext **k) {
-	if (k && *k)
-		g_option_context_free(*k);
-}
-static void free_gerror(GError **k) {
-	if (k && *k)
-		g_error_free(*k);
-}
+G_DEFINE_AUTOPTR_CLEANUP_FUNC(GOptionEntry, free)
 
 void config_load(int *argc, char ***argv, GOptionEntry *app_entries, const char *description,
 		char *default_config, char *default_section,
 		struct rtpengine_common_config *cconfig)
 {
-	AUTO_CLEANUP_NULL(GOptionContext *c, free_goptc);
-	AUTO_CLEANUP_NULL(GError *er, free_gerror);
-	AUTO_CLEANUP_GBUF(use_section);
+	g_autoptr(GOptionContext) c = NULL;
+	g_autoptr(GError) er = NULL;
+	g_autoptr(char) use_section = NULL;
 	const char *use_config;
 	int fatal = 0;
-	AUTO_CLEANUP(char **saved_argv, free_gvbuf) = g_strdupv(*argv);
+	g_autoptr(char_p) saved_argv = g_strdupv(*argv);
 	int saved_argc = *argc;
+	gboolean version = false;
 
 	rtpe_common_config_ptr = cconfig;
 
@@ -190,8 +175,9 @@ void config_load(int *argc, char ***argv, GOptionEntry *app_entries, const char
 #else
 	rtpe_common_config_ptr->default_log_level = LOG_DEBUG;
 #endif
+	rtpe_common_config_ptr->codec_chain_opus_complexity = -1;
 
-	AUTO_CLEANUP(GKeyFile *kf, free_gkeyfile) = g_key_file_new();
+	g_autoptr(GKeyFile) kf = g_key_file_new();
 
 #define ll(system, descr) \
 		{ "log-level-" #system,	0, 0, G_OPTION_ARG_INT,	&rtpe_common_config_ptr->log_levels[log_level_index_ ## system],"Log level for: " descr,"INT"		},
@@ -214,9 +200,17 @@ void config_load(int *argc, char ***argv, GOptionEntry *app_entries, const char
 		{ "foreground",		'f', 0, G_OPTION_ARG_NONE,	&rtpe_common_config_ptr->foreground,	"Don't fork to background",		NULL		},
 		{ "thread-stack",	0,0,	G_OPTION_ARG_INT,	&rtpe_common_config_ptr->thread_stack,	"Thread stack size in kB",		"INT"		},
 		{ "poller-size",	0,0,	G_OPTION_ARG_INT,	&rtpe_common_config_ptr->poller_size,	"Max poller items per iteration",	"INT"		},
+#ifdef HAVE_LIBURING
+		{ "io-uring",		0,0,	G_OPTION_ARG_NONE,	&rtpe_common_config_ptr->io_uring,	"Use io_uring",				NULL },
+		{ "io-uring-buffers",	0,0,	G_OPTION_ARG_INT,	&rtpe_common_config_ptr->io_uring_buffers,"Number of io_uring entries per thread","INT" },
+#endif
 		{ "evs-lib-path",	0,0,	G_OPTION_ARG_FILENAME,	&rtpe_common_config_ptr->evs_lib_path,	"Location of .so for 3GPP EVS codec",	"FILE"		},
 #ifdef HAVE_CODEC_CHAIN
 		{ "codec-chain-lib-path",0,0,	G_OPTION_ARG_FILENAME,	&rtpe_common_config_ptr->codec_chain_lib_path,"Location of libcodec-chain.so",	"FILE"		},
+		{ "codec-chain-runners",0,0,	G_OPTION_ARG_INT,	&rtpe_common_config_ptr->codec_chain_runners,"Number of chain runners per codec","INT"		},
+		{ "codec-chain-concurrency",0,0,G_OPTION_ARG_INT,	&rtpe_common_config_ptr->codec_chain_concurrency,"Max concurrent codec jobs per runner","INT"	},
+		{ "codec-chain-async",0,0,	G_OPTION_ARG_INT,	&rtpe_common_config_ptr->codec_chain_async,"Number of background callback threads","INT"	},
+		{ "codec-chain-opus-complexity",0,0,G_OPTION_ARG_INT,	&rtpe_common_config_ptr->codec_chain_opus_complexity,"Opus encoding complexity (0..10)","INT"	},
 #endif
 		{ NULL, }
 	};
@@ -227,10 +221,10 @@ void config_load(int *argc, char ***argv, GOptionEntry *app_entries, const char
 	unsigned int app_len = options_length(app_entries);
 	size_t entries_size = sizeof(GOptionEntry) * (shared_len + app_len + 1);
 
-	AUTO_CLEANUP(GOptionEntry *entries, free_gopte) = malloc(entries_size);
+	g_autoptr(GOptionEntry) entries = malloc(entries_size);
 	memcpy(entries, shared_options, sizeof(*entries) * shared_len);
 	memcpy(&entries[shared_len], app_entries, sizeof(*entries) * (app_len + 1));
-	AUTO_CLEANUP(GOptionEntry *entries_copy, free_gopte) = malloc(entries_size);
+	g_autoptr(GOptionEntry) entries_copy = malloc(entries_size);
 	memcpy(entries_copy, entries, entries_size);
 
 	c = g_option_context_new(description);
@@ -401,6 +395,29 @@ out:
 	if (rtpe_common_config_ptr->poller_size <= 0)
 		rtpe_common_config_ptr->poller_size = 128;
 
+#ifdef HAVE_CODEC_CHAIN
+	if (rtpe_common_config_ptr->codec_chain_runners <= 0)
+		rtpe_common_config_ptr->codec_chain_runners = 4;
+
+	if (rtpe_common_config_ptr->codec_chain_concurrency <= 0)
+		rtpe_common_config_ptr->codec_chain_concurrency = 256;
+
+	if (rtpe_common_config_ptr->codec_chain_async < 0)
+		rtpe_common_config_ptr->codec_chain_async = 0;
+
+	if (rtpe_common_config_ptr->codec_chain_opus_complexity == -1)
+		rtpe_common_config_ptr->codec_chain_opus_complexity = 10;
+	if (rtpe_common_config_ptr->codec_chain_opus_complexity < 0 || rtpe_common_config_ptr->codec_chain_opus_complexity > 10)
+		die("Invalid value for --codec-chain-opus-complexity");
+#endif
+
+#if HAVE_LIBURING
+	if (rtpe_common_config_ptr->io_uring_buffers == 0)
+		rtpe_common_config_ptr->io_uring_buffers = 16384;
+	else if (rtpe_common_config_ptr->io_uring_buffers < 0)
+		die("Invalid value for --io-uring-buffers");
+#endif
+
 	return;
 
 err:
@@ -465,29 +482,18 @@ int timeval_cmp_ptr(const void *a, const void *b) {
 	return 0;
 }
 
-void free_gbuf(char **p) {
-	g_free(*p);
-}
-
-void free_gvbuf(char ***p) {
-	g_strfreev(*p);
-}
-
-int g_tree_find_first_cmp(void *k, void *v, void *d) {
-	void **p = d;
-	GEqualFunc f = p[1];
-	if (!f || f(v, p[0])) {
-		p[2] = v;
+int rtpe_tree_find_first_cmp(void *k, void *v, void *d) {
+	struct rtpe_g_tree_find_helper *h = d;
+	if (!h->func || h->func(v, h->data)) {
+		h->out_p = v;
 		return TRUE;
 	}
 	return FALSE;
 }
-int g_tree_find_all_cmp(void *k, void *v, void *d) {
-	void **p = d;
-	GEqualFunc f = p[1];
-	GQueue *q = p[2];
-	if (!f || f(v, p[0]))
-		g_queue_push_tail(q, v);
+int rtpe_tree_find_all_cmp(void *k, void *v, void *d) {
+	struct rtpe_g_tree_find_helper *h = d;
+	if (!h->func || h->func(v, h->data))
+		g_queue_push_tail(h->out_q, v);
 	return FALSE;
 }
 
